from __future__ import annotations

from typing import TypeVar

import jax
import jax.numpy as jnp
from jax.dtypes import canonicalize_dtype

_T = TypeVar('_T', bound=tuple[None | jax.Array, ...])


def result_type(*args: None | jax.Array,
                dtype: jax.typing.DTypeLike | None = None,
                ensure_inexact: bool = True
                ) -> jax.typing.DTypeLike:
    """Find a common data type for arrays.

    Args:
        *args: Arrays to consider (or None).
        dtype: Overrides the inferred data type.
        ensure_inexact: Ensures that the result type is inexact, or raises.

    Returns:
        The common data type of *args.
    """
    if dtype is None:
        filtered_args = [x for x in args if x is not None]
        dtype = jnp.result_type(*filtered_args)
        if ensure_inexact and not jnp.issubdtype(dtype, jnp.inexact):
            dtype = jnp.promote_types(canonicalize_dtype(float), dtype)
    if ensure_inexact and not jnp.issubdtype(dtype, jnp.inexact):
        msg = f'Data type must be inexact: {dtype}'
        raise ValueError(msg)
    return dtype


def cast_to_result_type(arrays: _T,
                        /,
                        *,
                        dtype: None | jax.typing.DTypeLike = None,
                        ensure_inexact: bool = True
                        ) -> _T:
    """Casts the input arrays to a common data dtype.

    Args:
        arrays: Arrays to be converted (or None).
        dtype: Overrides the inferred data type.
        ensure_inexact: Ensures that the result type is inexact, or raises.

    Returns:
        The arrays cast to the inferred type.
    """
    dtype = result_type(*arrays, dtype=dtype, ensure_inexact=ensure_inexact)
    return tuple(jnp.asarray(x, dtype)  # type: ignore # pyright: ignore
                 if x is not None else None for x in arrays)



from math import prod

import numpy as np

from tjax import create_diagonal_array


def test_create_diagonal_array() -> None:
    s = (2, 3, 4)
    x = np.arange(prod(s), dtype=np.float64)
    y = np.reshape(x, s)
    z = create_diagonal_array(y)
    assert z.shape == (*s, s[-1])
    for index, value in np.ndenumerate(z):
        if index[-2] == index[-1]:
            y_index = index[:-1]
            assert y[y_index] == value
        else:
            assert value == 0



from __future__ import annotations

from functools import partial
from typing import Any

import jax.numpy as jnp
from jax import vjp
from numpy.testing import assert_equal

from tjax import assert_tree_allclose, copy_cotangent, cotangent_combinator, replace_cotangent


def test_copy_cotangent() -> None:
    primals, vjp_f = vjp(copy_cotangent, 1.0, 2.0)
    assert_equal(primals, 1.0)
    assert_equal(vjp_f(3.0), (3.0, 3.0))


def test_replace_cotangent() -> None:
    primals, vjp_f = vjp(replace_cotangent, 1.0, 2.0)
    assert_equal(primals, 1.0)
    assert_equal(vjp_f(3.0), (2.0, 3.0))


def test_combinator() -> None:
    def f(x: Any) -> tuple[Any, None]:
        return (x ** 2, x ** 2), None
    o = jnp.ones(())

    args = ((-1.0,), (-1.0,))
    _, f_vjp = vjp(partial(cotangent_combinator, f, aux_cotangent_scales=None), args)
    result_bar = (2 * o, 3 * o), None
    actual_args_bar, = f_vjp(result_bar)
    desired_args_bar = ((-4.0,), (-6.0,))
    assert_tree_allclose(actual_args_bar, desired_args_bar)



import jax.numpy as jnp
import numpy as np
import pytest
from jax import Array
from numpy.testing import assert_allclose

from tjax import JaxRealArray, leaky_integrate, leaky_integrate_time_series


def test_time_step_invariance() -> None:
    def f(n: int) -> JaxRealArray:
        x = jnp.asarray(0.0)
        for _ in range(n):
            x = leaky_integrate(x,
                                time_step=jnp.asarray(1 / n),
                                drift=jnp.asarray(3.0),
                                decay=jnp.asarray(2.0))
        assert isinstance(x, Array)
        return x

    a = np.asarray([f(n) for n in (1, 5, 20, 100)],
                   dtype=np.float64)
    assert_allclose(a, a[0])


@pytest.mark.parametrize('decay', [0.1, 1.0, 10.0])
def test_time_series_of_ones(decay: float) -> None:
    n = jnp.ones(100) * 12.0
    assert_allclose(n, leaky_integrate_time_series(n, jnp.asarray(decay)))


def test_time_series() -> None:
    time_series = jnp.asarray([0.1, 0.6, -0.3, 1.5])
    decay = jnp.asarray(0.3)
    result = jnp.asarray([0.1, 0.387221, 0.087076, 0.611119])
    assert_allclose(result, leaky_integrate_time_series(time_series, decay), rtol=3e-6)
